@@ -1,26 +1,29 @@
-# This file is part of the GNU Health GTK Client.  The COPYRIGHT file at the top level of this
+# This file is part of GNU Health.  The COPYRIGHT file at the top level of this
 # repository contains the full copyright notices and license terms.
 
 import operator
-import gobject
-import gtk
 import locale
 import decimal
 import gettext
 from decimal import Decimal
 
+from gi.repository import GLib, GObject, Gtk
+
 from .widget import Widget
-from tryton.config import CONFIG
+
 from tryton.gui.window.win_search import WinSearch
-from tryton.common import RPCExecute, RPCException, Tooltips, \
-    timezoned_date, untimezoned_date
-from tryton.common.placeholder_entry import PlaceholderEntry
+from tryton.common import Tooltips, timezoned_date, untimezoned_date, \
+        IconFactory
 from tryton.common.selection import selection_shortcuts
 from tryton.common.completion import get_completion, update_completion
 from tryton.common.datetime_ import Date, DateTime
 from tryton.common.domain_parser import quote
 from tryton.common.entry_position import reset_position
-from tryton.common.widget_style import set_widget_style
+from tryton.common.number_entry import NumberEntry
+from tryton.common.underline import set_underline
+from tryton.common.domain_inversion import eval_domain
+from tryton.common.widget_style import widget_class
+from tryton.pyson import PYSONDecoder
 
 _ = gettext.gettext
 
@@ -31,12 +34,12 @@ class DictEntry(object):
 
     def __init__(self, name, parent_widget):
         self.name = name
-        self.definition = parent_widget.keys[name]
+        self.definition = parent_widget.field.keys[name]
         self.parent_widget = parent_widget
         self.widget = self.create_widget()
 
     def create_widget(self):
-        widget = gtk.Entry()
+        widget = Gtk.Entry()
         widget.connect('key-press-event', self.parent_widget.send_modified)
         widget.connect('focus-out-event',
             lambda w, e: self.parent_widget._focus_out())
@@ -56,13 +59,12 @@ class DictEntry(object):
 
     def set_readonly(self, readonly):
         self.widget.set_editable(not readonly)
-        set_widget_style(self.widget, not readonly)
 
 
 class DictBooleanEntry(DictEntry):
 
     def create_widget(self):
-        widget = gtk.CheckButton()
+        widget = Gtk.CheckButton()
         widget.connect('toggled', self.parent_widget.sig_activate)
         widget.connect('focus-out-event', lambda w, e:
             self.parent_widget._focus_out())
@@ -88,7 +90,7 @@ class DictSelectionEntry(DictEntry):
     fill = False
 
     def create_widget(self):
-        widget = gtk.ComboBoxEntry()
+        widget = Gtk.ComboBox(has_entry=True)
 
         # customizing entry
         child = widget.get_child()
@@ -100,10 +102,13 @@ class DictSelectionEntry(DictEntry):
             lambda w: self.parent_widget._focus_out())
         widget.connect('notify::active',
             lambda w, e: self.parent_widget._focus_out())
+        widget.connect(
+            'scroll-event',
+            lambda c, e: c.stop_emission_by_name('scroll-event'))
         selection_shortcuts(widget)
 
         # setting completion and selection
-        model = gtk.ListStore(gobject.TYPE_STRING)
+        model = Gtk.ListStore(GObject.TYPE_STRING)
         model.append(('',))
         self._selection = {'': None}
         width = 10
@@ -116,9 +121,9 @@ class DictSelectionEntry(DictEntry):
             model.append((name,))
             width = max(width, len(name))
         widget.set_model(model)
-        widget.set_text_column(0)
+        widget.set_entry_text_column(0)
         child.set_width_chars(width)
-        completion = gtk.EntryCompletion()
+        completion = Gtk.EntryCompletion()
         completion.set_inline_selection(True)
         completion.set_model(model)
         child.set_completion(completion)
@@ -132,7 +137,7 @@ class DictSelectionEntry(DictEntry):
         text = child.get_text()
         value = None
         if text:
-            for txt, val in self._selection.items():
+            for txt, val in list(self._selection.items()):
                 if not val:
                     continue
                 if txt[:len(text)].lower() == text.lower():
@@ -156,32 +161,16 @@ class DictIntegerEntry(DictEntry):
     fill = False
 
     def create_widget(self):
-        widget = super(DictIntegerEntry, self).create_widget()
-        widget.set_width_chars(8)
-        widget.set_max_length(0)
-        widget.set_alignment(1.0)
-        widget.connect('insert-text', self.sig_insert_text)
+        widget = NumberEntry()
+        widget.connect('key-press-event', self.parent_widget.send_modified)
+        widget.connect('focus-out-event',
+            lambda w, e: self.parent_widget._focus_out())
+        widget.props.activates_default = True
+        widget.connect('activate', self.parent_widget.sig_activate)
         return widget
 
-    def sig_insert_text(self, entry, new_text, new_text_length, position):
-        value = entry.get_text()
-        position = entry.get_position()
-        new_value = value[:position] + new_text + value[position:]
-        if new_value == '-':
-            return
-        try:
-            locale.atoi(new_value)
-        except ValueError:
-            entry.stop_emission('insert-text')
-
     def get_value(self):
-        txt_value = self.widget.get_text()
-        if txt_value:
-            try:
-                return locale.atoi(txt_value)
-            except ValueError:
-                pass
-        return None
+        return int(self.widget.value)
 
     def set_value(self, value):
         if value is not None:
@@ -194,57 +183,37 @@ class DictIntegerEntry(DictEntry):
 
 class DictFloatEntry(DictIntegerEntry):
 
+    @property
     def digits(self):
-        default = (16, 2)
         record = self.parent_widget.record
-        if not record:
-            return default
-        return tuple(y if x is None else x for x, y in zip(
-                record.expr_eval(self.definition.get('digits', default)),
-                default))
+        if record:
+            digits = record.expr_eval(self.definition.get('digits'))
+            if not digits or any(d is None for d in digits):
+                return
+            return digits
 
-    def sig_insert_text(self, entry, new_text, new_text_length, position):
-        value = entry.get_text()
-        position = entry.get_position()
-        new_value = value[:position] + new_text + value[position:]
-        decimal_point = locale.localeconv()['decimal_point']
-
-        if new_value in ('-', decimal_point):
-            return
-
-        digits = self.digits()
-
-        try:
-            locale.atof(new_value)
-        except ValueError:
-            entry.stop_emission('insert-text')
-            return
-
-        new_int = new_value
-        new_decimal = ''
-        if decimal_point in new_value:
-            new_int, new_decimal = new_value.rsplit(decimal_point, 1)
-
-        if (len(new_int) > digits[0]
-                or len(new_decimal) > digits[1]):
-            entry.stop_emission('insert-text')
+    @property
+    def width(self):
+        digits = self.digits
+        if digits:
+            return sum(digits)
+        else:
+            return 18
 
     def get_value(self):
-        txt_value = self.widget.get_text()
-        if txt_value:
-            try:
-                return locale.atof(txt_value)
-            except ValueError:
-                pass
-        return None
+        return self.widget.value
 
     def set_value(self, value):
-        digits = self.digits()
+        digits = self.digits
+        if digits:
+            self.widget.digits = digits[1]
+        else:
+            self.widget.digits = None
+        self.widget.set_width_chars(self.width)
         if value is not None:
-            txt_val = locale.format('%.' + str(digits[1]) + 'f', value, True)
+            txt_val = locale.format('%.*f', (digits[1], value), True)
         else:
             txt_val = ''
-        self.widget.set_width_chars(sum(digits))
         self.widget.set_text(txt_val)
         reset_position(self.widget)
 
@@ -324,31 +293,30 @@ class DictWidget(Widget):
     def __init__(self, view, attrs):
         super(DictWidget, self).__init__(view, attrs)
         self.schema_model = attrs['schema_model']
-        self.keys = {}
         self.fields = {}
         self.buttons = {}
         self.rows = {}
 
-        self.widget = gtk.Frame()
-        self.widget.set_label(attrs.get('string', ''))
-        self.widget.set_shadow_type(gtk.SHADOW_OUT)
+        self.widget = Gtk.Frame()
+        label = Gtk.Label(label=set_underline(attrs.get('string', '')))
+        label.set_use_underline(True)
+        self.widget.set_label_widget(label)
+        self.widget.set_shadow_type(Gtk.ShadowType.OUT)
 
-        vbox = gtk.VBox()
+        vbox = Gtk.VBox()
         self.widget.add(vbox)
 
-        self.table = gtk.Table(1, 3, homogeneous=False)
-        self.table.set_col_spacings(0)
-        self.table.set_row_spacings(0)
-        self.table.set_border_width(0)
-        vbox.pack_start(self.table, expand=True, fill=True)
+        self.grid = Gtk.Grid(column_spacing=3, row_spacing=3)
+        vbox.pack_start(self.grid, expand=True, fill=True, padding=0)
 
-        hbox = gtk.HBox()
+        hbox = Gtk.HBox()
         hbox.set_border_width(2)
-        self.wid_text = PlaceholderEntry()
+        self.wid_text = Gtk.Entry()
         self.wid_text.set_placeholder_text(_('Search'))
         self.wid_text.props.width_chars = 13
         self.wid_text.connect('activate', self._sig_activate)
-        hbox.pack_start(self.wid_text, expand=True, fill=True)
+        hbox.pack_start(self.wid_text, expand=True, fill=True, padding=0)
+        label.set_mnemonic_widget(self.wid_text)
 
         if int(self.attrs.get('completion', 1)):
             self.wid_completion = get_completion(search=False, create=False)
@@ -359,16 +327,13 @@ class DictWidget(Widget):
         else:
             self.wid_completion = None
 
-        self.but_add = gtk.Button()
+        self.but_add = Gtk.Button(can_focus=False)
         self.but_add.connect('clicked', self._sig_add)
-        img_add = gtk.Image()
-        img_add.set_from_stock('tryton-list-add', gtk.ICON_SIZE_SMALL_TOOLBAR)
-        img_add.set_alignment(0.5, 0.5)
-        self.but_add.add(img_add)
-        self.but_add.set_relief(gtk.RELIEF_NONE)
-        hbox.pack_start(self.but_add, expand=False, fill=False)
-        hbox.set_focus_chain([self.wid_text])
-        vbox.pack_start(hbox, expand=True, fill=True)
+        self.but_add.add(
+            IconFactory.get_image('tryton-add', Gtk.IconSize.SMALL_TOOLBAR))
+        self.but_add.set_relief(Gtk.ReliefStyle.NONE)
+        hbox.pack_start(self.but_add, expand=False, fill=False, padding=0)
+        vbox.pack_start(hbox, expand=True, fill=True, padding=0)
 
         self.tooltips = Tooltips()
         self.tooltips.set_tip(self.but_add, _('Add value'))
@@ -377,14 +342,15 @@ class DictWidget(Widget):
         self._readonly = False
         self._record_id = None
 
+    @property
+    def _invalid_widget(self):
+        return self.wid_text
+
     def _new_remove_btn(self):
-        but_remove = gtk.Button()
-        img_remove = gtk.Image()
-        img_remove.set_from_stock('tryton-list-remove',
-            gtk.ICON_SIZE_SMALL_TOOLBAR)
-        img_remove.set_alignment(0.5, 0.5)
-        but_remove.add(img_remove)
-        but_remove.set_relief(gtk.RELIEF_NONE)
+        but_remove = Gtk.Button()
+        but_remove.add(
+            IconFactory.get_image('tryton-remove', Gtk.IconSize.SMALL_TOOLBAR))
+        but_remove.set_relief(Gtk.ReliefStyle.NONE)
         return but_remove
 
     def _sig_activate(self, *args):
@@ -393,7 +359,7 @@ class DictWidget(Widget):
 
     def _sig_add(self, *args):
         context = self.field.get_context(self.record)
-        value = self.wid_text.get_text().decode('utf-8')
+        value = self.wid_text.get_text()
         domain = self.field.domain_get(self.record)
 
         def callback(result):
@@ -407,55 +373,48 @@ class DictWidget(Widget):
         win.show()
 
     def add_new_keys(self, ids):
-        context = self.field.get_context(self.record)
+        new_keys = self.field.add_new_keys(ids, self.record)
         self.send_modified()
-        try:
-            new_fields = RPCExecute('model', self.schema_model,
-                'get_keys', ids, context=context)
-        except RPCException:
-            new_fields = []
         focus = False
-        for new_field in new_fields:
-            if new_field['name'] not in self.fields:
-                self.keys[new_field['name']] = new_field
-                self.add_line(new_field['name'])
+        for key_name in new_keys:
+            if key_name not in self.fields:
+                self.add_line(key_name)
                 if not focus:
                     # Use idle add because it can be called from the callback
                     # of WinSearch while the popup is still there
-                    gobject.idle_add(
-                        self.fields[new_field['name']].widget.grab_focus)
+                    GLib.idle_add(self.fields[key_name].widget.grab_focus)
                     focus = True
 
     def _sig_remove(self, button, key, modified=True):
         del self.fields[key]
         del self.buttons[key]
         for widget in self.rows[key]:
-            self.table.remove(widget)
+            self.grid.remove(widget)
             widget.destroy()
         del self.rows[key]
         if modified:
             self.send_modified()
-            self.set_value(self.record, self.field)
+            self.set_value()
 
-    def set_value(self, record, field):
-        field.set_client(record, self.get_value())
+    def set_value(self):
+        self.field.set_client(self.record, self.get_value())
 
     def get_value(self):
         return dict((key, widget.get_value())
-            for key, widget in self.fields.items())
+            for key, widget in list(self.fields.items()))
 
     @property
     def modified(self):
         if self.record and self.field:
             value = self.field.get_client(self.record)
             return any(widget.modified(value)
-                for widget in self.fields.itervalues())
+                for widget in self.fields.values())
         return False
 
     def _readonly_set(self, readonly):
         self._readonly = readonly
         self._set_button_sensitive()
-        for widget in self.fields.values():
+        for widget in list(self.fields.values()):
             widget.set_readonly(readonly)
         self.wid_text.set_sensitive(not readonly)
         self.wid_text.set_editable(not readonly)
@@ -464,91 +423,68 @@ class DictWidget(Widget):
         self.but_add.set_sensitive(bool(
                 not self._readonly
                 and self.attrs.get('create', True)))
-        for button in self.buttons.itervalues():
+        for button in self.buttons.values():
             button.set_sensitive(bool(
                     not self._readonly
                     and self.attrs.get('delete', True)))
 
     def add_line(self, key):
-        self.fields[key] = DICT_ENTRIES[self.keys[key]['type_']](key, self)
+        key_schema = self.field.keys[key]
+        self.fields[key] = DICT_ENTRIES[key_schema['type_']](key, self)
         field = self.fields[key]
-        alignment = gtk.Alignment(
-            float(self.attrs.get('xalign', 0.0)),
-            float(self.attrs.get('yalign', 0.5)),
-            float(self.attrs.get('xexpand', 1.0)),
-            float(self.attrs.get('yexpand', 1.0)))
-        hbox = gtk.HBox()
-        hbox.pack_start(field.widget, expand=field.expand, fill=field.fill)
-        alignment.add(hbox)
-        n_rows = self.table.props.n_rows
-        self.table.resize(n_rows + 1, 3)
-        text = self.keys[key]['string'] + _(':')
-        label = gtk.Label(text)
-        label.set_alignment(1., .5)
-        self.table.attach(label, 0, 1, n_rows - 1, n_rows,
-            xoptions=gtk.FILL, yoptions=False, xpadding=2)
+        text = key_schema['string'] + _(':')
+        label = Gtk.Label(
+            label=set_underline(text),
+            use_underline=True, halign=Gtk.Align.END)
+        self.grid.attach_next_to(
+            label, None, Gtk.PositionType.BOTTOM, 1, 1)
         label.set_mnemonic_widget(field.widget)
         label.show()
-        self.table.attach(alignment, 1, 2, n_rows - 1, n_rows,
-            xoptions=gtk.FILL | gtk.EXPAND, yoptions=False, xpadding=2)
-        alignment.show_all()
+        hbox = Gtk.HBox(hexpand=True)
+        hbox.pack_start(
+            field.widget, expand=field.expand, fill=field.fill, padding=0)
+        self.grid.attach_next_to(
+            hbox, label, Gtk.PositionType.RIGHT, 1, 1)
+        hbox.show_all()
         remove_but = self._new_remove_btn()
         self.tooltips.set_tip(remove_but, _('Remove "%s"') %
-            self.keys[key]['string'])
-        self.table.attach(remove_but, 2, 3, n_rows - 1, n_rows,
-            xoptions=gtk.FILL, yoptions=False, xpadding=2)
+            key_schema['string'])
+        self.grid.attach_next_to(
+            remove_but, hbox, Gtk.PositionType.RIGHT, 1, 1)
         remove_but.connect('clicked', self._sig_remove, key)
         remove_but.show_all()
-        self.rows[key] = [label, alignment, remove_but]
+        self.rows[key] = [label, hbox, remove_but]
         self.buttons[key] = remove_but
 
-    def add_keys(self, keys):
-        context = self.field.get_context(self.record)
-        domain = self.field.domain_get(self.record)
-        batchlen = min(10, CONFIG['client.limit'])
-        for i in xrange(0, len(keys), batchlen):
-            sub_keys = keys[i:i + batchlen]
-            try:
-                key_ids = RPCExecute('model', self.schema_model, 'search',
-                    [('name', 'in', sub_keys), domain], 0,
-                    CONFIG['client.limit'], None, context=context)
-            except RPCException:
-                key_ids = []
-            if not key_ids:
-                continue
-            try:
-                values = RPCExecute('model', self.schema_model,
-                    'get_keys', key_ids, context=context)
-            except RPCException:
-                values = []
-            if not values:
-                continue
-            self.keys.update({k['name']: k for k in values})
+    def display(self):
+        super(DictWidget, self).display()
 
-    def display(self, record, field):
-        super(DictWidget, self).display(record, field)
-
-        if field is None:
+        if not self.field:
             return
 
-        record_id = record.id if record else None
+        record_id = self.record.id if self.record else None
         if record_id != self._record_id:
-            for key in self.fields.keys():
+            for key in list(self.fields.keys()):
                 self._sig_remove(None, key, modified=False)
             self._record_id = record_id
 
-        value = field.get_client(record) if field else {}
-        new_key_names = set(value.iterkeys()) - set(self.keys)
+        value = self.field.get_client(self.record) if self.field else {}
+        new_key_names = set(value.keys()) - set(self.field.keys)
         if new_key_names:
-            self.add_keys(list(new_key_names))
-        for key, val in sorted(value.iteritems()):
-            if key not in self.keys:
+            self.field.add_keys(list(new_key_names), self.record)
+        decoder = PYSONDecoder()
+        for key, val in sorted(value.items()):
+            if key not in self.field.keys:
                 continue
             if key not in self.fields:
                 self.add_line(key)
             widget = self.fields[key]
             widget.set_value(val)
             widget.set_readonly(self._readonly)
+            key_domain = decoder.decode(
+                self.field.keys[key].get('domain') or '[]')
+            widget_class(
+                widget.widget, 'invalid', not eval_domain(key_domain, value))
         for key in set(self.fields.keys()) - set(value.keys()):
             self._sig_remove(None, key, modified=False)
 
